/**
  Tiny OpenGL/GLUT particle system example program
  
  This version attaches springs between selected particles.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-01-21 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */

#include "ogl/glsl.h" /* programmable shaders */
#include "ogl/glsl.cpp"

#include "ogl/minicam.h" /* small camera control functions */

#include <vector> /* std::vector container */

/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

class particle {
public:
	vec3 P; /* position */
	vec3 V; /* velocity */
	vec3 F; /* temporary summed-up force from all my springs */
	vec3 C; /* color */
};

class spring {
public:
	float k; /* converts deformation to force */
	float restlen; /* length of string at rest */
	int a,b; /* numbers of my particles */
};


/* Return the distance between these two pointers, in bytes,
itself expressed as a pointer.  Needed for glVertexPointer &c. */
void *ptr_diff(void *a,void *b) {
	return (void *)(long)((char *)a - (char *)b);
}

/* Make an OpenGL texture handle, of this size, with this pixel format */
GLuint make_texture(int w,int h,GLenum format=GL_RGBA8)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,tex);
	char *mem=new char[w*h*4]; memset(mem,0x80,w*h*4);
	glTexImage2D(t,0,format, w,h,0, GL_RGBA,GL_UNSIGNED_BYTE,mem);
	delete[] mem;
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,GL_LINEAR); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR); 
	
	glTexParameterf(t, GL_TEXTURE_MIN_LOD, 0.0);
	glGenerateMipmapEXT(t);
	
	GLenum texWrap=GL_CLAMP_TO_EDGE; /* stop changing at edge */
	glTexParameteri(t, GL_TEXTURE_WRAP_S, texWrap); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, texWrap);
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_texture");
	return tex;
}

/* Make a framebuffer object that renders pixels into this texture */
GLuint make_fbo(GLuint tex)
{
	GLuint fb=0;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
	
	/* Make sure the framebuffer's happy */
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
	{ fprintf(stderr,"Error %d in framebuffer creation!\n",status); exit(1); }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after make_fbo");
	return fb;
}

/* Erase this framebuffer object to this color */
void erase_fbo(GLuint fb,vec4 color) {
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glClearColor(color.x,color.y,color.z,color.w);
	glClear(GL_COLOR_BUFFER_BIT);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after erase_fbo");
}

void display(void) /* Draw everything in our window */
{
	unsigned int i;
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float dt=0.001, display_dt=0.0;
	static float sim_time=0.0;
	float big=30; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
/* Particle state storage */
	static bool first_time=true; /* for initial bc's! */
	static std::vector<particle> pts;
	static std::vector<spring> springs;
	if (first_time) { /* First time: initialize points */
		static int count=0;
		srand(count++);
		for (int i=0;i<100;i++) {
			particle p;
			p.P=0.3*rv(); 
			p.V=0.3*(rv()-vec3(0.5)); 
			p.C=rv();
			pts.push_back(p);
			for (int interconnect=0;interconnect<i;interconnect++)
			{
				spring s;
				s.k=10.0;
				s.a=interconnect; s.b=i;
				s.restlen=length(pts[s.a].P-pts[s.b].P);
				springs.push_back(s);
			}
		}
		
		first_time=true;
	}
	
	static float energy=0, m=1.0; /* mass and energy (for energy accounting) */
	
	
	/* Subcycle, so the physics runs at the right stepsize, 
	   yet still keeps up with the display. */
	double wall_time=oglTimer(); //  *0.1; <- for slow-motion 
	while (sim_time<wall_time && !key_down[(int)' '])
	{ /* take one physics step */
		sim_time+=dt;
		energy=0;
		
		/* zero net force */
		for (i=0;i<pts.size();i++) pts[i].F=vec3(0.0);
		
		if (key_down[(int)'x']) pts[0].F.x+=10.0;
		if (key_down[(int)'X']) pts[0].F.x-=10.0;
		if (key_down[(int)'y']) pts[0].F.y+=10.0;
		if (key_down[(int)'Y']) pts[0].F.y-=10.0;
		if (key_down[(int)'z']) pts[0].F.z+=50.0;
		
		/* physics the springs to find the net force on all points */
		for (i=0;i<springs.size();i++) {
			spring &s=springs[i];
			vec3 dir=pts[s.b].P-pts[s.a].P;
			float len=length(dir);
			float x=len-s.restlen;
			float force=-s.k*x;
			vec3 force3=force*normalize(dir);
			pts[s.a].F-=force3;
			pts[s.b].F+=force3;
		}
		
		/* physics the particles */
		for (i=0;i<pts.size();i++) {
			particle &p=pts[i];

		/* Apply physics */
			vec3 gA(0.0,0.0,-0.1); /* gravity acceleration */

			float KE=0.5*m*dot(p.V,p.V);
			float PE=m*dot(gA,-p.P);
			energy+=PE+KE;
			
			float m=1.0;
			vec3 fF=-0.5*p.V; /* frictional force */
			vec3 A=(fF+p.F)/m+gA; /* total acceleration */

			p.P+=dt*p.V+0.5*dt*dt*A; /* move position by velocity */
			p.V+=dt*A; /* accelerate particles downwards */
			
			float r=0.1;
			vec3 center(0.5,0.5,0.05);
			vec3 d=p.P-center;
			float pr=length(d);
			if (pr<r) { /* our point is inside the sphere--bounce off */
				/* Reflect velocity around d vector: */
				vec3 nd=d*(1.0/pr); /* normalized d vector == sphere normal (facing outwards) */
				p.V=p.V-2.0*nd*dot(p.V,nd);
				if (first_time) p.P=center+d*(r/pr); /* scale to surface */
			}
			
			/* Bounce off all six sides of cube */
			for (int dir=-1;dir<=1;dir+=2)
			for (int axis=0;axis<3;axis++) {
				float loc=(p.P[axis]-0.5)*2.0;
				if (loc*dir>1.0 && p.V[axis]*dir>0.0) { /* bounce off this axis */
					p.V[axis]*=-1.0;
				}
			}
		}

		first_time=false;
	}
	
	
/* Shadow texture */
	static int dots_pix=1024; /* pixels in our little dots texture */
	static GLuint dots_tex=make_texture(dots_pix,dots_pix,GL_RGBA8);
	static GLuint dots_fbo=make_fbo(dots_tex);
	static int fbo_setup=1;
	if (fbo_setup) erase_fbo(dots_fbo,vec4(1,0,0,1)); /* opaque red */
	fbo_setup=0;
	
	/* Bind the dots texture framebuffer, so we can draw our little dots drops */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, dots_fbo);
	glViewport(0,0,dots_pix,dots_pix);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	/* From 0..1 texture coords to -1..+1 clip coords */
	glTranslatef(-1.0,-1.0,-1.0);
	glScalef(2.0,2.0,2.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glEnable(GL_BLEND);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(20.0);
	
	/* Erase background */
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1);
	glVertex2f(0,0); glVertex2f(1,0); glVertex2f(1,1); glVertex2f(0,1);
	glEnd();
	
	/* Draw black splatty point shadows */
	glBegin(GL_POINTS);
	glColor4f(0,0,0,1.0); 
	for (i=0;i<pts.size();i++) {
		particle &p=pts[i];
		glVertex3f(p.P.x,p.P.y,0);
	}
	glEnd();
	

/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.001,1000.0 /* Z range */
	);
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Draw in a ground plane
	glDisable(GL_DEPTH_TEST); /* don't read/write from depth buffer */
	//glDepthMask(GL_FALSE); /* don't write to depth buffer (no Z fighting!) */
	
	glBegin(GL_QUADS);
	glColor4f(0.6,0.4,0.2,1.0); /* light, opaque brown */
	glVertex3f(-big,-big,floor);
	glVertex3f(+big,-big,floor);
	glVertex3f(+big,+big,floor);
	glVertex3f(-big,+big,floor);
	glEnd();
	
	// Draw our dots texture on the ground
	glActiveTextureARB(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,dots_tex);
	glGenerateMipmapEXT(GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	float sz=1.0;
	glBegin(GL_QUADS);
	glTexCoord2fv(vec3(0,0,0)); glVertex3f(0,0,floor);
	glTexCoord2fv(vec3(1,0,0)); glVertex3f(sz,0,floor);
	glTexCoord2fv(vec3(1,1,0)); glVertex3f(sz,sz,floor);
	glTexCoord2fv(vec3(0,1,0)); glVertex3f(0,sz,floor);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	// Draw in a 1-meter grid on the ground
	glEnable(GL_LINE_SMOOTH);
	glBegin(GL_LINES);
	glColor4f(0.0f,0.0f,0.0f,0.3f);
	for (int v=-big;v<=big;v++) {
		glVertex3f(-big,v,floor); /* X lines */
		glVertex3f(+big,v,floor);
		glVertex3f(v,-big,floor); /* Y lines */
		glVertex3f(v,+big,floor);
	}
	glEnd();
	
	glDepthMask(GL_TRUE); /* depth writes back on */
	
/* Draw axes */
	oglCameraAxes();
	
	glBegin(GL_LINES);
	for (i=0;i<springs.size();i++) {
		spring &s=springs[i];
	glColor3f(0.5,0.5,0.5);
		glVertex3fv(pts[s.a].P);
		glVertex3fv(pts[s.b].P);
	glColor4f(0,0,0,0.3);
		glVertex2fv(pts[s.a].P);
		glVertex2fv(pts[s.b].P);
	}
	glEnd();
	
/* Draw our points */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(8.4);
	/* Make gl_PointSize work from GLSL */
        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	
	static GLhandleARB simplePoint=makeProgramObject(
	"varying vec4 drawColor;\n"
	"uniform float pixPerParticle;\n"
	"void main(void) {\n"
	"	drawColor = gl_Color;\n"
	"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
	"	gl_PointSize = pixPerParticle/gl_Position.w;\n"
	"}\n"
	,
	"varying vec4 drawColor;\n"
	"void main(void) {\n"
	"	gl_FragColor = drawColor; \n"
	"}\n"
	);
	glUseProgramObjectARB(simplePoint);
	glUniform1fARB(glGetUniformLocationARB(simplePoint,"pixPerParticle"),
		14.0e-3*glutGet(GLUT_WINDOW_HEIGHT));
	
	glBegin(GL_POINTS);
	for (i=0;i<pts.size();i++) {
		particle &p=pts[i];
	/* Draw the particle */
		glColor3fv(p.C);
		glVertex3fv(p.P);
	}
	glEnd();
	glUseProgramObjectARB(0); /* back to plain non-programmable OpenGL */
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed;
		printf("energy = %g\n",energy);
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	camera=vec3(0.5,0.5,1.0);
	camera_orient.y=vec3(0,1,0); camera_orient.z=vec3(0,0,1);

	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutCreateWindow("Spring system demo (try WASD and XYZ keys!)"); /* make a window */
	glewInit(); /* only AFTER create window! */
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	glutMainLoop(); /* start drawing */
}
